"""Builds automatic documentation of the installed webviz-config plugins.
The documentation is designed to be used by the YAML configuration file end
user. Sphinx has not been used, as the documentation from Sphinx is geared
mostly towards Python end users. It is also a small task generating `webviz`
documentation, and most of the Sphinx machinery is not needed.

Overall workflow is:
    * Find all installed plugins.
    * Automatically read docstring and __init__ function signatures (both
      argument names and which arguments have default values).
    * Output the extracted plugin information into docsify input using jinja2.
"""

import shutil
import inspect
import pathlib
from importlib import import_module
from collections import defaultdict
from typing import Any, Dict, Optional, Tuple, List

try:
    # Python 3.8+
    # pylint: disable=ungrouped-imports
    from importlib.metadata import version  # type: ignore
    from typing import TypedDict  # type: ignore
except (ImportError, ModuleNotFoundError):
    # Python < 3.8
    from importlib_metadata import version  # type: ignore
    from typing_extensions import TypedDict  # type: ignore

import jinja2

import webviz_config.plugins
from webviz_config._config_parser import SPECIAL_ARGS


class ArgInfo(TypedDict, total=False):
    required: bool
    default: Any
    typehint: Any
    typehint_string: str


class PluginInfo(TypedDict):
    arg_info: Dict[str, ArgInfo]
    argument_description: Optional[str]
    data_input: Optional[str]
    description: Optional[str]
    module: str
    name: str
    package: str
    package_doc: Optional[str]
    package_version: str


def _document_plugin(plugin: Tuple[str, Any]) -> PluginInfo:
    """Takes in a tuple (from e.g. inspect.getmembers), and returns
    a dictionary according to the type definition PluginInfo.
    """

    name, reference = plugin
    docstring = reference.__doc__ if reference.__doc__ is not None else ""
    docstring_parts = _split_docstring(docstring)
    argspec = inspect.getfullargspec(reference.__init__)
    module = inspect.getmodule(reference)
    subpackage = inspect.getmodule(module).__package__  # type: ignore
    top_package_name = subpackage.split(".")[0]  # type: ignore

    plugin_info: PluginInfo = {
        "arg_info": {arg: {} for arg in argspec.args if arg not in SPECIAL_ARGS},
        "argument_description": docstring_parts[1]
        if len(docstring_parts) > 1
        else None,
        "data_input": docstring_parts[2] if len(docstring_parts) > 2 else None,
        "description": docstring_parts[0] if docstring != "" else None,
        "name": name,
        "module": module.__name__,  # type: ignore
        "package": top_package_name,
        "package_doc": import_module(subpackage).__doc__,  # type: ignore
        "package_version": version(top_package_name),
    }

    if argspec.defaults is not None:
        for arg, default in dict(
            zip(reversed(argspec.args), reversed(argspec.defaults))
        ).items():
            plugin_info["arg_info"][arg]["default"] = default

    for arg, arg_info in plugin_info["arg_info"].items():
        arg_info["required"] = "default" not in arg_info

    for arg, annotation in argspec.annotations.items():
        if arg not in SPECIAL_ARGS:
            plugin_info["arg_info"][arg]["typehint"] = annotation
            plugin_info["arg_info"][arg]["typehint_string"] = _annotation_to_string(
                annotation
            )

    return plugin_info


def get_plugin_documentation() -> defaultdict:
    """Find all installed webviz plugins, and then document them
    by grabbing docstring and input arguments / function signature.
    """

    plugin_doc = [
        _document_plugin(plugin)
        for plugin in inspect.getmembers(webviz_config.plugins, inspect.isclass)
        if not plugin[0].startswith("Example")
    ]

    # Sort the plugins by package:
    package_ordered: defaultdict = defaultdict(lambda: {"plugins": []})
    for sorted_plugin in sorted(plugin_doc, key=lambda x: (x["module"], x["name"])):
        package = sorted_plugin["package"]
        package_ordered[package]["plugins"].append(sorted_plugin)
        package_ordered[package]["doc"] = sorted_plugin["package_doc"]
        package_ordered[package]["version"] = sorted_plugin["package_version"]

    return package_ordered


def _annotation_to_string(annotation: Any) -> str:
    """Takes in a type annotation (that could come from e.g. inspect.getfullargspec)
    and transforms it into a human readable string.
    """

    def remove_fix(string: str, fix: str, prefix: bool = True) -> str:
        if prefix and string.startswith(fix):
            return string[len(fix) :]
        if not prefix and string.endswith(fix):
            return string[: -len(fix)]
        return string

    text_type = str(annotation)
    text_type = remove_fix(text_type, "typing.")
    text_type = remove_fix(text_type, "<class '")
    text_type = remove_fix(text_type, "'>", prefix=False)
    text_type = text_type.replace("pathlib.Path", "str (corresponding to a path)")

    return text_type


def build_docs(build_directory: pathlib.Path) -> None:

    # From Python 3.8, copytree gets an argument dirs_exist_ok.
    # Then the rmtree command can be removed.
    shutil.rmtree(build_directory)
    shutil.copytree(
        pathlib.Path(__file__).resolve().parent / "static",
        build_directory,
    )

    # While waiting for https://github.com/jhildenbiddle/docsify-tabs/pull/30
    # to be reviewed, merged and released:
    substring_prefix = (
        r"/[\r\n]*(\s*)<!-+\s+tab:\s*(.*)\s+-+>[\r\n]+([\s\S]*?)[\r\n]*\s*"
    )
    docsify_tabs_file = build_directory / "docsify-tabs.min.js"
    docsify_tabs_file.write_text(
        docsify_tabs_file.read_text().replace(
            substring_prefix + r"(?=<!-+\s+tabs?:)/m",
            substring_prefix + r"(?=<!-+\s+(tab:\s*(.*)|tabs:\s*?end)\s+-+>)/m",
        )
    )

    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )

    plugin_documentation = get_plugin_documentation()

    template = template_environment.get_template("README.md.jinja2")
    for package_name, package_doc in plugin_documentation.items():
        (build_directory / f"{package_name}.md").write_text(
            template.render({"package_name": package_name, "package_doc": package_doc})
        )

    template = template_environment.get_template("sidebar.md.jinja2")
    (build_directory / "sidebar.md").write_text(
        template.render({"packages": plugin_documentation.keys()})
    )

    template = template_environment.get_template("webviz-doc.js.jinja2")
    (build_directory / "webviz-doc.js").write_text(
        template.render(
            {
                "paths": ["/"]
                + [f"/{package_name}" for package_name in plugin_documentation]
            }
        )
    )


def _split_docstring(docstring: str) -> List[str]:
    """Divides docstring by splitting on ---, also unindents
    first in case of indented docstrings (similar to this one)
    """
    lines = docstring.strip().split("\n")

    try:
        indent_spaces = min(
            [len(line) - len(line.lstrip()) for line in lines[1:] if line.strip() != ""]
        )
    except ValueError:  # In the case of no original newline (\n)
        indent_spaces = 0

    unindented_lines = [lines[0]] + [line[indent_spaces:] for line in lines[1:]]
    return "\n".join(unindented_lines).split("\n---\n")



import os
import atexit
import pathlib
import shutil
import tempfile

from .._is_reload_process import is_reload_process
from ._certificate_generator import (
    create_certificate,
    SERVER_KEY_FILENAME,
    SERVER_CRT_FILENAME,
)


class LocalhostCertificate:
    """Facilitates creating certificates only valid for localhost, on the fly,
    when starting a local process. The certificates are stored in a temporary folder
    only readable by the user running the process, and are deleted on exit.
    """

    def __init__(self) -> None:
        if not is_reload_process():
            self._ssl_temp_dir = pathlib.Path(tempfile.mkdtemp())
            os.environ["WEBVIZ_SSL_TEMP_DIR"] = str(self._ssl_temp_dir)
            create_certificate(self._ssl_temp_dir)
            atexit.register(self._delete_temp_dir)
        else:
            self._ssl_temp_dir = pathlib.Path(os.environ["WEBVIZ_SSL_TEMP_DIR"])

    def _delete_temp_dir(self) -> None:
        """Delete temporary directory with on-the-fly generated localhost certificates"""
        shutil.rmtree(self._ssl_temp_dir)

    @property
    def ssl_context(self) -> tuple:
        return (
            self._ssl_temp_dir / SERVER_CRT_FILENAME,
            self._ssl_temp_dir / SERVER_KEY_FILENAME,
        )



"""These are the basic Webviz configuration plugins, distributed through
the utility itself.
"""

try:
    # Python 3.8+
    from importlib.metadata import entry_points  # type: ignore
except ModuleNotFoundError:
    # Python < 3.8
    from importlib_metadata import entry_points  # type: ignore

from ._example_plugin import ExamplePlugin
from ._example_data_download import ExampleDataDownload
from ._example_assets import ExampleAssets
from ._example_portable import ExamplePortable
from ._example_tour import ExampleTour
from ._banner_image import BannerImage
from ._data_table import DataTable
from ._syntax_highlighter import SyntaxHighlighter
from ._table_plotter import TablePlotter
from ._embed_pdf import EmbedPdf
from ._markdown import Markdown

__all__ = [
    "ExamplePlugin",
    "ExampleAssets",
    "ExamplePortable",
    "ExampleTour",
    "BannerImage",
    "DataTable",
    "SyntaxHighlighter",
    "TablePlotter",
    "EmbedPdf",
    "Markdown",
]

for entry_point in entry_points().get("webviz_config_plugins", []):
    globals()[entry_point.name] = entry_point.load()
    __all__.append(entry_point.name)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

TESTS_REQUIRES = [
    "bandit",
    "black",
    "jsonschema",
    "mock",
    "mypy",
    "pylint~=2.3",
    "pytest-xdist",
    "selenium~=3.141",
]

setup(
    name="webviz-config",
    description="Configuration file support for webviz",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/equinor/webviz-config",
    author="R&T Equinor",
    packages=find_packages(exclude=["tests"]),
    package_data={
        "webviz_config": [
            "_docs/static/*",
            "_docs/static/fonts/*",
            "static/*",
            "static/.dockerignore",
            "static/assets/*",
            "templates/*",
            "themes/default_assets/*",
        ]
    },
    entry_points={"console_scripts": ["webviz=webviz_config.command_line:main"]},
    install_requires=[
        "bleach>=3.1",
        "cryptography>=2.4",
        "dash>=1.16",
        "flask-caching>=1.4",
        "flask-talisman>=0.6",
        "jinja2>=2.10",
        "markdown>=3.0",
        "pandas>=0.24",
        "pyarrow>=0.16",
        "pyyaml>=5.1",
        "tqdm>=4.8",
        "importlib-metadata>=1.7; python_version<'3.8'",
        "typing-extensions>=3.7; python_version<'3.8'",
        "webviz-core-components>=0.1.0",
    ],
    tests_require=TESTS_REQUIRES,
    extras_require={"tests": TESTS_REQUIRES},
    setup_requires=["setuptools_scm~=3.2"],
    python_requires="~=3.6",
    use_scm_version=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Environment :: Web Environment",
        "Framework :: Dash",
        "Framework :: Flask",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: MIT License",
    ],
)



import re
import sys
import pathlib
import inspect
import importlib
import typing
import types

import yaml

from . import plugins as standard_plugins
from . import WebvizPluginABC
from .utils import terminal_colors

SPECIAL_ARGS = ["self", "app", "_call_signature", "_imports"]


def _get_webviz_plugins(module: types.ModuleType) -> list:
    """Returns a list of all Webviz plugins
    in the module given as input.
    """

    def _is_webviz_plugin(obj: typing.Any) -> bool:
        return inspect.isclass(obj) and issubclass(obj, WebvizPluginABC)

    return [member[0] for member in inspect.getmembers(module, _is_webviz_plugin)]


def _call_signature(
    module: types.ModuleType,
    module_name: str,
    plugin_name: str,
    kwargs: dict,
    config_folder: pathlib.Path,
    contact_person: typing.Optional[dict] = None,
) -> tuple:
    # pylint: disable=too-many-branches,too-many-statements
    """Takes as input the name of a plugin, the module it is located in,
    together with user given arguments (originating from the configuration
    file). Returns the equivalent Python code wrt. initiating an instance of
    that plugin (with the given arguments).

    Raises ParserError in the following scenarios:
      * User is missing a required (i.e. no default value) __init__ argument
      * User provided an argument not existing in the class' __init__ function
      * User has given one of the protected arguments in SPECIAL_ARGS
      * If there is type mismatch between user given argument value, and type
        hint in __init__ signature (given that type hint exist)
    """
    argspec = inspect.getfullargspec(getattr(module, plugin_name).__init__)

    if argspec.defaults is not None:
        required_args = argspec.args[: -len(argspec.defaults)]
    else:
        required_args = argspec.args

    for arg in required_args:
        if arg not in SPECIAL_ARGS and arg not in kwargs:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"`{plugin_name}` requires the argument "
                f"`{arg}` in your configuration file."
                f"{terminal_colors.END}"
            )

    for arg in list(kwargs):
        if arg in SPECIAL_ARGS:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"Argument `{arg}` not allowed."
                f"{terminal_colors.END}"
            )

        if arg == "contact_person":
            if not isinstance(kwargs["contact_person"], dict):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"The contact information provided for "
                    f"`{plugin_name}` is not a dictionary. "
                    f"{terminal_colors.END}"
                )

            if any(
                key not in ["name", "phone", "email"]
                for key in kwargs["contact_person"]
            ):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"Unrecognized contact information key given to `{plugin_name}`. "
                    f"Should be 'name', 'phone' and/or 'email'."
                    f"{terminal_colors.END}"
                )

            contact_person = kwargs.pop("contact_person")

        elif arg not in argspec.args:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"Unrecognized argument. `{plugin_name}` "
                f"does not take an argument `{arg}`."
                f"{terminal_colors.END}"
            )

        if arg in argspec.annotations:
            expected_type = argspec.annotations[arg]

            if expected_type == pathlib.Path:
                kwargs[arg] = (config_folder / pathlib.Path(kwargs[arg])).resolve()
            elif expected_type == typing.List[pathlib.Path]:
                kwargs[arg] = [
                    (config_folder / pathlib.Path(patharg)).resolve()
                    for patharg in kwargs[arg]
                ]
            try:
                if not isinstance(kwargs[arg], expected_type):
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        f"The value provided for argument `{arg}` "
                        f"given to `{plugin_name}` is "
                        f"of type `{type(kwargs[arg]).__name__}`. "
                        f"Expected type "
                        f"`{argspec.annotations[arg].__name__}`."
                        f"{terminal_colors.END}"
                    )
            # Typechecking typing classes does not work in python 3.7
            except TypeError:
                pass

    special_args = ""
    if "app" in argspec.args:
        special_args += "app=app, "

    return (
        f"{module_name}.{plugin_name}({special_args}**{kwargs})",
        f"plugin_layout(contact_person={contact_person})",
    )


class ParserError(Exception):
    pass


class ConfigParser:

    STANDARD_PLUGINS = _get_webviz_plugins(standard_plugins)

    def __init__(self, yaml_file: pathlib.Path):

        ConfigParser.check_for_tabs_in_file(yaml_file)

        try:
            self._configuration = yaml.safe_load(yaml_file.read_text())
        except yaml.MarkedYAMLError as excep:
            extra_info = (
                f"There is something wrong in the configuration file {yaml_file}. "
            )

            if hasattr(excep, "problem_mark"):
                extra_info += (
                    "The typo is probably somewhere around "
                    f"line {excep.problem_mark.line + 1}."
                )

            raise type(excep)(
                f"{excep}. {terminal_colors.RED}{terminal_colors.BOLD}"
                f"{extra_info}{terminal_colors.END}"
            ).with_traceback(sys.exc_info()[2])

        self._config_folder = pathlib.Path(yaml_file).parent
        self._page_ids: typing.List[str] = []
        self._assets: set = set()
        self.clean_configuration()

    @staticmethod
    def check_for_tabs_in_file(path: pathlib.Path) -> None:

        line_numbers_with_tabs = sorted(
            list(
                {
                    i + 1
                    for i, line in enumerate(path.read_text().splitlines())
                    if "\t" in line
                }
            )
        )

        if line_numbers_with_tabs:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration file contains tabs. You should use ordinary spaces "
                "instead. The following lines contain tabs:\n"
                f"{', '.join([str(line_no) for line_no in line_numbers_with_tabs])}"
                f"{terminal_colors.END}"
            )

    def _generate_page_id(self, title: str) -> str:
        """From the user given title, this function provides a unique
        human readable page id, not already present in self._page_ids
        """

        base_id = re.sub("[^-a-z0-9_]+", "", title.lower().replace(" ", "_"))

        page_id = base_id

        count = 1
        while page_id in self._page_ids:
            count += 1
            page_id = f"{base_id}{count}"

        return page_id

    def clean_configuration(self) -> None:
        # pylint: disable=too-many-branches,too-many-statements
        """Various cleaning and checks of the raw configuration read
        from the user provided yaml configuration file.
        """

        self.configuration["_imports"] = set()

        if "shared_settings" in self.configuration:
            self._shared_settings = self.configuration["shared_settings"]
        else:
            self._shared_settings = {}

        if "title" not in self.configuration:
            self.configuration["title"] = "Webviz - Powered by Dash"

        if "pages" not in self.configuration:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration file does not have "
                "information regarding which pages to create."
                f"{terminal_colors.END}"
            )

        if not isinstance(self.configuration["pages"], list):
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration input belonging to the "
                "`pages` keyword should be a list."
                f"{terminal_colors.END}"
            )

        for page_number, page in enumerate(self.configuration["pages"]):

            if "title" not in page:
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"Page number {page_number + 1} does "
                    "not have the title specified."
                    f"{terminal_colors.END}"
                )

            if "id" not in page:
                page["id"] = self._generate_page_id(page["title"])
            elif page["id"] in self._page_ids:
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    "You have more than one page "
                    "with the same `id`."
                    f"{terminal_colors.END}"
                )

            self._page_ids.append(page["id"])

            if "content" not in page:
                page["content"] = []
            elif not isinstance(page["content"], list):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    "The content of page number "
                    f"{page_number + 1} should be a list."
                    f"{terminal_colors.END}"
                )

            plugins = [e for e in page["content"] if isinstance(e, dict)]

            for plugin in plugins:
                plugin_name = next(iter(plugin))
                plugin_variables = next(iter(plugin.values()))
                kwargs = {} if plugin_variables is None else {**plugin_variables}

                if "." not in plugin_name:
                    if plugin_name not in ConfigParser.STANDARD_PLUGINS:
                        raise ParserError(
                            f"{terminal_colors.RED}{terminal_colors.BOLD}"
                            "You have included a plugin with "
                            f"name `{plugin_name}` in your "
                            "configuration file. This is not a "
                            "standard plugin."
                            f"{terminal_colors.END}"
                        )

                    self.configuration["_imports"].add(
                        ("webviz_config.plugins", "standard_plugins")
                    )

                    plugin["_call_signature"] = _call_signature(
                        standard_plugins,
                        "standard_plugins",
                        plugin_name,
                        kwargs,
                        self._config_folder,
                    )

                    self.assets.update(getattr(standard_plugins, plugin_name).ASSETS)

                else:
                    parts = plugin_name.split(".")

                    plugin_name = parts[-1]
                    module_name = ".".join(parts[:-1])
                    module = importlib.import_module(module_name)

                    if plugin_name not in _get_webviz_plugins(module):
                        raise ParserError(
                            f"{terminal_colors.RED}{terminal_colors.BOLD}"
                            f"Module `{module}` does not have a "
                            f"plugin named `{plugin_name}`"
                            f"{terminal_colors.END}"
                        )

                    self.configuration["_imports"].add(module_name)
                    plugin["_call_signature"] = _call_signature(
                        module,
                        module_name,
                        plugin_name,
                        kwargs,
                        self._config_folder,
                    )

                    self.assets.update(getattr(module, plugin_name).ASSETS)

    @property
    def configuration(self) -> dict:
        return self._configuration

    @property
    def shared_settings(self) -> dict:
        return self._shared_settings

    @property
    def assets(self) -> set:
        return self._assets



import sys
import time
import shutil
import pathlib
import tempfile
import subprocess  # nosec
import argparse

from yaml import YAMLError

from ._config_parser import ParserError
from ._write_script import write_script
from .themes import installed_themes
from .utils import terminal_colors

BUILD_FILENAME = "webviz_app.py"
STATIC_FOLDER = pathlib.Path(__file__).resolve().parent / "static"


def build_webviz(args: argparse.Namespace) -> None:

    if args.theme not in installed_themes:
        raise ValueError(f"Theme `{args.theme}` is not installed.")

    if args.portable is None:
        build_directory = pathlib.Path(tempfile.mkdtemp())
    else:
        build_directory = args.portable.resolve()
        build_directory.mkdir(parents=True)

    shutil.copytree(STATIC_FOLDER / "assets", build_directory / "assets")

    for filename in ["README.md", "Dockerfile", ".dockerignore"]:
        shutil.copy(STATIC_FOLDER / filename, build_directory)

    for asset in installed_themes[args.theme].assets:
        shutil.copy(asset, build_directory / "assets")

    (build_directory / "theme_settings.json").write_text(
        installed_themes[args.theme].to_json()
    )

    try:
        if args.portable:
            print(
                f"{terminal_colors.BLUE}{terminal_colors.BOLD}"
                "Saving requested data to build folder "
                "such that the webviz instance is portable."
                f"{terminal_colors.END}"
            )

            write_script(
                args, build_directory, "copy_data_template.py.jinja2", "copy_data.py"
            )

            if subprocess.call(  # nosec
                [sys.executable, "copy_data.py"], cwd=build_directory
            ):
                raise RuntimeError(
                    "Something went wrong. This is probably "
                    "not related to webviz-config, but more "
                    "likely in one of the  dependencies used "
                    "by one or more containers. See exact "
                    "error message and traceback above"
                )

            (build_directory / "copy_data.py").unlink()

            print(
                f"{terminal_colors.GREEN}{terminal_colors.BOLD}"
                "Finished data extraction. All output saved."
                f"{terminal_colors.END}"
            )

        non_default_assets = write_script(
            args, build_directory, "webviz_template.py.jinja2", BUILD_FILENAME
        )

        for asset in non_default_assets:
            shutil.copy(asset, build_directory / "assets")

        if not args.portable:
            run_webviz(args, build_directory)

    finally:
        if not args.portable:
            shutil.rmtree(build_directory)


def run_webviz(args: argparse.Namespace, build_directory: pathlib.Path) -> None:

    print(
        f"{terminal_colors.YELLOW}"
        " Starting up your webviz application. Please wait..."
        f"{terminal_colors.END}"
    )

    app_process = subprocess.Popen(  # nosec
        [sys.executable, BUILD_FILENAME], cwd=build_directory
    )

    lastmtime = args.yaml_file.stat().st_mtime

    while app_process.poll() is None:
        try:
            time.sleep(1)

            if lastmtime != args.yaml_file.stat().st_mtime:
                lastmtime = args.yaml_file.stat().st_mtime
                write_script(
                    args, build_directory, "webviz_template.py.jinja2", BUILD_FILENAME
                )
                print(
                    f"{terminal_colors.BLUE}{terminal_colors.BOLD}"
                    " Rebuilt webviz dash app from configuration file"
                    f"{terminal_colors.END}"
                )

        except (ParserError, YAMLError) as excep:
            print(
                f"{excep} {terminal_colors.RED}{terminal_colors.BOLD}"
                "Fix the error and save the configuration file in "
                " order to trigger a new rebuild."
                f"{terminal_colors.END}"
            )

        except KeyboardInterrupt:
            app_process.kill()
            print(
                f"\r{terminal_colors.BLUE}{terminal_colors.BOLD}"
                " Shutting down the webviz application on user request."
                f"{terminal_colors.END}"
            )
            app_process.wait()

        except Exception as excep:
            app_process.kill()
            print(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "Unexpected error. Killing the webviz application process."
                f"{terminal_colors.END}"
            )
            app_process.wait()
            raise excep
